from __future__ import annotations

import os
from loguru import logger
from .base import *
from .filesystem_base import Minio_CloudFileSystem
from .filesystem_pathlib import *


class FileMinioPurePath(CloudFileSystemPurePath):
    _prefix: str = 'minio'
    _posix_prefix: str = 's3'
    _provider: str = 'Minio'
    _win_pathz: ClassVar = 'PureFileMinioWindowsPath'
    _posix_pathz: ClassVar = 'PureFileMinioPosixPath'


class PureFileMinioPosixPath(PureCloudFileSystemPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PureFileMinioWindowsPath(PureCloudFileSystemWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()

class FileMinioPath(CloudFileSystemPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 'minio'
    _posix_prefix = 's3'
    _provider = 'Minio'

    _win_pathz: ModuleType = 'FileMinioWindowsPath'
    _posix_pathz: ModuleType = 'FileMinioPosixPath'

    def _init(self, template: Optional['FileMinioPath'] = None):
        self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is FileMinioPath or issubclass(cls, FileMinioPath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self

    @property
    def _cloudpath(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        #if self._prefix in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        #return self._pathlike.join(*self.parts)
        if self._prefix in self.parts[0] or self._posix_prefix in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        return self._pathlike.join(*self.parts)
    
    @property
    def _s3_cloudstr(self) -> str:
        """
        Reconstructs the proper cloud URI
        """
        if self._prefix not in self.parts[0] and self._posix_prefix not in self.parts[0]:
            return f'{self._posix_prefix}://' + '/'.join(self.parts)
        return f'{self._posix_prefix}://' + '/'.join(self.parts[1:])

    
    def glob(self, pattern: str = '*', as_path: bool = True) -> Iterable[Union[str, Type['CloudFileSystemPath']]]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        Warning: doesn't work as expected. Use Find Instead.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))
        #if self.is_cloud:
        glob_pattern = self._s3_cloudstr + ('/' if self.is_dir and not self._path.endswith('/') and not pattern.startswith('/') else '') +  pattern
        try: 
            matches =  self._accessor.glob(glob_pattern)
            if not matches: return matches
            if self.is_cloud: matches = [f'{self._prefix}://{m}' for m in matches]
            if as_path: matches = [type(self)(m) for m in matches]
            return matches
        except Exception as e: 
            logger.error(e)
            return self.find(pattern = pattern, as_string = not as_path)

    async def async_glob(self, pattern: str = '*', as_path: bool = True) -> AsyncIterable[Type['CloudFileSystemPath']]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))
        glob_pattern = self._s3_cloudstr + ('/' if self.is_dir and not self._path.endswith('/') and not pattern.startswith('/') else '') +  pattern
        try: 
            matches = await self._accessor.async_glob(glob_pattern)
            #logger.info(glob_pattern)
            if not matches: return matches
            if self.is_cloud: matches = [f'{self._prefix}://{m}' for m in matches]
            if as_path: matches = [type(self)(m) for m in matches]
            return matches
        except Exception as e: 
            logger.error(e)
            return await self.async_find(pattern = pattern, as_string = not as_path)
    
    def find(self, pattern: str = "*",  as_string: bool = False, maxdepth: int = None, withdirs: bool = None, detail: bool = False) -> Union[List[str], List[Type['CloudFileSystemPath']]]:
        """
        List all files below path. Like posix find command without conditions
        """
        matches = self._accessor.find(path = self._s3_cloudstr, maxdepth = maxdepth, withdirs = withdirs, detail = detail, prefix = pattern)
        if self.is_cloud:
            matches = [f'{self._prefix}://{m}' for m in matches]
        if not as_string:
            matches = [type(self)(m) for m in matches]
        return matches
    
    async def async_find(self, pattern: str = "*",  as_string: bool = False, maxdepth: int = None, withdirs: bool = None, detail: bool = False) -> Union[List[str], List[Type['CloudFileSystemPath']]]:
        """
        List all files below path. Like posix find command without conditions
        """
        matches = await self._accessor.async_find(path = self._s3_cloudstr, maxdepth = maxdepth, withdirs = withdirs, detail = detail, prefix = pattern)
        if self.is_cloud:
            matches = [f'{self._prefix}://{m}' for m in matches]
        if not as_string:
            matches = [type(self)(m) for m in matches]
        return matches


class FileMinioPosixPath(PosixPath, FileMinioPath, PureFileMinioPosixPath):
    __slots__ = ()


class FileMinioWindowsPath(WindowsPath, FileMinioPath, PureFileMinioWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FileMinioPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FileMinioPath.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        FileMinioPurePath, FileMinioPath, PureFileMinioPosixPath, FileMinioWindowsPath, FileMinioPosixPath, PureFileMinioWindowsPath
    ]
)

MinioFileSystem = Minio_CloudFileSystem


__all__ = (
    'FileMinioPurePath',
    'FileMinioPath',
    'PureFileMinioPosixPath',
    'FileMinioWindowsPath',
    'FileMinioPosixPath',
    'PureFileMinioWindowsPath',
    'MinioFileSystem'
)



"""
Handles Importing the correct pathlib for this library in
python3.10 cases
"""

import sys
import os
import io
import inspect
import ntpath
import posixpath

from anyio import open_file, AsyncFile
from os import stat_result, PathLike
from contextlib import asynccontextmanager

from typing import TYPE_CHECKING, Union
from typing import Optional, List, AsyncIterable, Iterable, IO, AsyncContextManager, cast, Callable
from .aiopath.wrap import coro_as_method_coro, func_as_method_coro, to_thread, method_as_method_coro, func_to_async_func
from .aiopath.handle import IterableAIOFile, get_handle
from .aiopath.types import Final, Literal, FileMode


# if 3.10
if sys.version_info.minor >= 10:
    from . import _pathlib as pathlib
    from ._pathlib import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from ._pathlib import _NormalAccessor as NormalAccessor
    from ._pathlib import _make_selector as _sync_make_selector
    from ._pathlib import _PosixFlavour, _WindowsFlavour
    try:
        from ._pathlib import _getfinalpathname
        _async_getfinalpathname = func_to_async_func(_getfinalpathname)

    except ImportError:
        def _getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")

        async def _async_getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")
        
else:
    import pathlib
    from pathlib import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from pathlib import _NormalAccessor as NormalAccessor
    from pathlib import _make_selector as _sync_make_selector
    from pathlib import _PosixFlavour, _WindowsFlavour
    try:
        from pathlib import _getfinalpathname
        _async_getfinalpathname = func_to_async_func(_getfinalpathname)

    except ImportError:
        def _getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")
    
        async def _async_getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")



DEFAULT_ENCODING: Final[str] = 'utf-8'
ON_ERRORS: Final[str] = 'ignore'
NEWLINE: Final[str] = '\n'

BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = '\n'
ENCODING: Final[str] = 'utf-8'
ERRORS: Final[str] = 'replace'

FileData = Union[bytes, str]
Paths = Union[Path, PathLike, str]
Handle = AsyncFile


def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    if hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__): return True
    return False


__all__ = (
    'io',
    'os',
    'inspect',
    'ntpath',
    'posixpath',
    'open_file',
    'AsyncFile',
    'stat_result',
    'PathLike',
    'asynccontextmanager',
    'pathlib', 
    'PosixPath',
    'WindowsPath',
    'PurePath',
    'Path',
    '_ignore_error',
    '_PosixFlavour',
    '_WindowsFlavour',
    'NormalAccessor',
    '_sync_make_selector',
    '_getfinalpathname',
    '_async_getfinalpathname',
    'coro_as_method_coro',
    'func_as_method_coro',
    'to_thread',
    'method_as_method_coro',
    'func_to_async_func',
    'IterableAIOFile',
    'Final',
    'Literal',
    'get_handle',
    'FileMode',
    'DEFAULT_ENCODING',
    'ON_ERRORS',
    'NEWLINE',
    'TYPE_CHECKING',
    'BEGINNING',
    'CHUNK_SIZE',
    'SEP',
    'ENCODING',
    'ERRORS',
    'FileData', 'Paths', 'Handle',
    'iscoroutinefunction',
    'Union', 'Optional', 'List', 'AsyncIterable', 'Iterable', 'IO', 'AsyncContextManager', 'cast', 'Callable'
)


from __future__ import annotations
import os
from .base import *

from loguru import logger
from .filesystem_base import AWS_CloudFileSystem
from .filesystem_pathlib import *


class FileS3PurePath(CloudFileSystemPurePath):
    _prefix: str = 's3'
    _provider: str = 'AmazonS3'
    _win_pathz: ClassVar = 'PureFileS3WindowsPath'
    _posix_pathz: ClassVar = 'PureFileS3PosixPath'


class PureFileS3PosixPath(PureCloudFileSystemPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PureFileS3WindowsPath(PureCloudFileSystemWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()

class FileS3Path(CloudFileSystemPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 's3'
    _provider = 'AmazonS3'

    _win_pathz: ModuleType = 'FileS3WindowsPath'
    _posix_pathz: ModuleType = 'FileS3PosixPath'

    def _init(self, template: Optional['FileS3Path'] = None):
        self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is FileS3Path or issubclass(cls, FileS3Path): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self
    
    # Implement some stuff that boto is faster in
    def upload_file(self, dest: PathLike, filename: Optional[str] = None, overwrite: bool = True, **kwargs):
        """
        Upload a file to S3

        Utilize boto3
        """
        if not overwrite and dest.exists(): raise FileExistsError(f"{dest} already exists and overwrite is False")
        filename = filename or self.name
        self._accessor.boto.upload_file(
            Bucket = self._bucket,
            Key = self.get_path_key(filename),
            Filename = dest.as_posix(),
        )
        return self.parent.joinpath(filename)
    
    def download_file(
        self, 
        output_file: Optional[PathLike] = None,
        output_dir: Optional[PathLike] = None,
        filename: Optional[str] = None,
        overwrite: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
        ):
        """
        Downloads a file from S3 to a path
        """
        assert output_file or output_dir, "Must provide either output_file or output_dir"
        output_file = output_file or output_dir.joinpath(filename or self.name)
        assert overwrite or not output_file.exists(), f"{output_file} already exists and overwrite is False"
        s3t = self._accessor.s3t()
        s3t.download(
            self._bucket,
            self.get_path_key(self.name),
            output_file.as_posix(),
            subscribers = callbacks
        )
        s3t.shutdown()
        return output_file

    async def async_upload_file(self, dest: PathLike, filename: Optional[str] = None,  overwrite: bool = True, **kwargs):
        """
        Upload a file to S3

        Utilize boto3
        """
        if not overwrite and await dest.async_exists(): raise FileExistsError(f"{dest} already exists and overwrite is False")
        filename = filename or self.name
        s3t = self._accessor.s3t()
        s3t.upload(
            dest.as_posix(),
            self._bucket,
            self.get_path_key(filename)
        )
        s3t.shutdown()
        #await to_thread(
        #    self._accessor.boto.upload_file, Bucket = self._bucket, Key = self.get_path_key(filename), Filename = dest.as_posix()
        #)
        return self.parent.joinpath(filename)

    def batch_upload_files(
        self, 
        files: Optional[List[PathLike]] = None,
        glob_path: Optional[str] = None,
        overwrite: bool = False,
        skip_existing: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
    ):
        """
        Handles batch uploading of files

        https://stackoverflow.com/questions/56639630/how-can-i-increase-my-aws-s3-upload-speed-when-using-boto3
        """
        assert files or glob_path, "Must provide either files or glob_path"
        if glob_path: files = list(self.glob(glob_path))
        results = []
        s3t = self._accessor.s3t()
        for file in files:
            if not overwrite and skip_existing and file.exists(): continue
            s3t.upload(
                file.as_posix(),
                self._bucket,
                self.get_path_key(file.name),
                subscribers = callbacks
            )
            results.append(self.parent.joinpath(file.name))
        s3t.shutdown()
        return results
    
    def batch_download_files(
        self,
        glob_path: str,
        output_dir: PathLike,
        overwrite: bool = False,
        skip_existing: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
    ):
        """
        Handles batch downloading of files
        """
        files = list(self.glob(glob_path))
        results = []
        s3t = self._accessor.s3t()
        for file in files:
            if not overwrite and skip_existing and file.exists(): continue
            output_file = output_dir.joinpath(file.name)
            s3t.download(
                self._bucket,
                self.get_path_key(file.name),
                output_file.as_posix(),
                subscribers = callbacks
            )
            results.append(output_file)
        s3t.shutdown()
        return results



class FileS3PosixPath(PosixPath, FileS3Path, PureFileS3PosixPath):
    __slots__ = ()


class FileS3WindowsPath(WindowsPath, FileS3Path, PureFileS3WindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FileS3Path.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FileS3Path.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        FileS3PurePath, FileS3Path, PureFileS3PosixPath, FileS3WindowsPath, FileS3PosixPath, PureFileS3WindowsPath
    ]
)

AWSFileSystem = AWS_CloudFileSystem


__all__ = (
    'FileS3PurePath',
    'FileS3Path',
    'PureFileS3PosixPath',
    'FileS3WindowsPath',
    'FileS3PosixPath',
    'PureFileS3WindowsPath',
    'AWSFileSystem'
)



from __future__ import annotations

import os
from .base import *
from .filesystem_base import GCP_CloudFileSystem
from .filesystem_pathlib import *


class FileGSPurePath(CloudFileSystemPurePath):
    _prefix: str = 'gs'
    _provider: str = 'GoogleCloudStorage'
    _win_pathz: ClassVar = 'PureFileGSWindowsPath'
    _posix_pathz: ClassVar = 'PureFileGSPosixPath'


class PureFileGSPosixPath(PureCloudFileSystemPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()
    _prefix = 'gs'
    _provider = 'GoogleCloudStorage'


class PureFileGSWindowsPath(PureCloudFileSystemWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()
    _prefix = 'gs'
    _provider = 'GoogleCloudStorage'

class FileGSPath(CloudFileSystemPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 'gs'
    _provider = 'GoogleCloudStorage'

    _win_pathz: ModuleType = 'FileGSWindowsPath'
    _posix_pathz: ModuleType = 'FileGSPosixPath'

    def _init(self, template: Optional['FileGSPath'] = None):
        self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is FileGSPath or issubclass(cls, FileGSPath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self


class FileGSPosixPath(PosixPath, FileGSPath, PureFileGSPosixPath):
    __slots__ = ()


class FileGSWindowsPath(WindowsPath, FileGSPath, PureFileGSWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FileGSPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FileGSPath.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        FileGSPurePath, FileGSPath, PureFileGSPosixPath, FileGSWindowsPath, FileGSPosixPath, PureFileGSWindowsPath
    ]
)

GCPFileSystem = GCP_CloudFileSystem


__all__ = (
    'FileGSPurePath',
    'FileGSPath',
    'PureFileGSPosixPath',
    'FileGSWindowsPath',
    'FileGSPosixPath',
    'PureFileGSWindowsPath',
    'GCPFileSystem'
)



from typing import Optional, List, Union, AsyncIterable

try:
  from typing import Literal, Final

except ImportError:
  from typing_extensions import Literal, Final


TextMode = Literal['r', 'w', 'a', 'x', 'r+', 'w+', 'a+', 'x+']
BinaryMode = Literal['rb', 'wb', 'ab', 'xb', 'r+b', 'w+b', 'a+b', 'x+b']
FileMode = Union[TextMode, BinaryMode]


from __future__ import annotations


from ..base_imports import *
from ..flavours import _pathz_windows_flavour, _pathz_posix_flavour

from fileio.types import *
from .methods import _ASYNC_SYNTAX_MAPPING

if TYPE_CHECKING:
    from fileio.base import FilePath

URI_PREFIXES = ('gs://', 's3://', 'minio://', 'mio://', 's3compat://')
_URI_SCHEMES = frozenset(('gs', 's3', 'minio', 'mio', 's3compat'))
_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
    'mio://': '/minio/',
    'minio://': '/minio/',
    's3compat://': '/s3compat/'
}
_PROVIDER_MAP = {
    'gs': 'GoogleCloudStorage',
    's3': 'AmazonS3',
    'mio': 'MinIO',
    'minio': 'MinIO',
    's3compat': 'S3Compatible'
}

Paths = Union['FilePath', Path, str]


def rewrite_async_syntax(obj, provider: str = 's3'):
    """
    Basically - we're rewriting all the fsspec's async method
    from _method to async_method for syntax
    """
    _names = _ASYNC_SYNTAX_MAPPING[provider]
    for attr in dir(obj):
        if attr.startswith('_') and not attr.startswith('__'):
            attr_val = getattr(obj, attr)
            if iscoroutinefunction(attr_val) and _names.get(attr):
                setattr(obj, _names[attr], attr_val)
    return obj

@asynccontextmanager
async def get_cloud_handle(name: Paths, mode: FileMode = 'r', buffering: int = -1, encoding: str | None = ENCODING, errors: str | None = ERRORS, newline: str | None = SEP) -> AsyncContextManager[Handle]:
    file: AsyncFile
    if 'b' in mode: file = await open_file(name, mode)
    else: file = await open_file(name, mode, encoding=encoding, errors=errors, newline=newline)
    yield file
    await file.aclose()


@asynccontextmanager
async def get_cloud_file(filelike: Paths) -> AsyncContextManager[Handle]:
    file: AsyncFile
    filelike = cast(IO[Union[str, bytes, os.PathLike]], filelike)
    file = AsyncFile(filelike)
    yield file
    await file.aclose()




from types import ModuleType
from typing import Callable, Any, Optional, Coroutine, Type, Union, List, ClassVar

try: import gcsfs
except ImportError: gcsfs: ModuleType = None

try: import s3fs
except ImportError: s3fs: ModuleType = None

try: import boto3
except ImportError: boto3: ModuleType = None

#try: import botocore
#except ImportError: botocore: ModuleType = None

from .base import rewrite_async_syntax, NormalAccessor, func_as_method_coro


class CloudFileSystemType(type):
    fs: ModuleType = None
    fsa: ModuleType = None
    fs_name: str = None # gcsfs
    boto: ModuleType = None
    s3t: Callable = None
    #s3t: 'boto3.s3.transfer.TransferManager' = None

    @classmethod
    def is_ready(cls):
        return bool(cls.fsa and cls.fs)
    
    #@classmethod
    def build_gcsfs(cls, **auth_config):
        from .._modules import LibModule
        from .config import CloudConfig

        gcsfs: ModuleType = LibModule.import_lib('gcsfs')
        LibModule.reload_module(gcsfs)

        authz = CloudConfig()
        if auth_config: authz.update_auth(**auth_config)
        _config = {}
        if authz.google_application_credentials.exists: _config['token'] = authz.google_application_credentials.as_posix()
        gcp_project = authz.get_gcp_project()
        if gcp_project: _config['project'] = gcp_project
        if authz.gcs_client_config: _config['client_kwargs'] = authz.gcs_client_config
        if authz.gcs_config: _config['config_kwargs'] = authz.gcs_config
        cls.fs = gcsfs.GCSFileSystem(asynchronous=False, **_config)
        cls.fsa = rewrite_async_syntax(gcsfs.GCSFileSystem(asynchronous=True, **_config), 'gs')
    
    #@classmethod
    def build_s3fs(cls, **auth_config):
        from .._modules import LibModule
        from .config import CloudConfig

        s3fs = LibModule.import_lib('s3fs')
        LibModule.reload_module(s3fs)
        
        boto3 = LibModule.import_lib('boto3')
        LibModule.reload_module(boto3)

        authz = CloudConfig()
        if auth_config: authz.update_auth(**auth_config)
        _config = {}
        if authz.aws_access_key_id:
            _config['key'] = authz.aws_access_key_id
            _config['secret'] = authz.aws_secret_access_key
        elif authz.aws_access_token: _config['token'] = authz.aws_access_token
        elif not authz.boto_config: _config['anon'] = True
        if authz.s3_config: _config['config_kwargs'] = authz.s3_config
        cls.fs = s3fs.S3FileSystem(asynchronous=False, **_config)
        cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **_config))
        
        from botocore.config import Config as BotoConfig
        import boto3.s3.transfer as s3transfer
        boto_config = BotoConfig(max_pool_connections = authz.max_workers)
        cls.boto = boto3.client('s3', region_name = authz.aws_region, config = boto_config)
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = authz.max_workers,
        )
        def create_s3t(self):
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        #cls.s3t = lambda: s3transfer.create_transfer_manager(cls.boto, transfer_config)
        cls.s3t = create_s3t
        #cls.s3t = s3transfer.create_transfer_manager(cls.boto, transfer_config)
    
    #classmethod
    def build_minio(cls, **auth_config):
        from .._modules import LibModule
        from .config import CloudConfig

        s3fs: ModuleType = LibModule.import_lib('s3fs')
        LibModule.reload_module(s3fs)

        boto3 = LibModule.import_lib('boto3')
        LibModule.reload_module(boto3)
        
        authz = CloudConfig()
        if auth_config: authz.update_auth(**auth_config)
        _config = {}        
        
        if authz.minio_secret_key:
            _config['key'] = authz.minio_access_key
            _config['secret'] = authz.minio_secret_key
        elif authz.minio_access_token: _config['token'] = authz.minio_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.minio_endpoint, 'region_name': authz.aws_region}
        _config['config_kwargs'] = authz.minio_config or {}
        _config['config_kwargs']['signature_version'] = authz.minio_signature_ver
        cls.fs = s3fs.S3FileSystem(asynchronous=False, **_config)
        cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **_config))
        
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig
        
        boto_config = BotoConfig(
            signature_version = authz.minio_signature_ver,
            max_pool_connections = authz.max_workers
        )
        cls.boto = boto3.client('s3', region_name = authz.aws_region, endpoint_url = authz.minio_endpoint, aws_access_key_id = authz.minio_access_key, aws_secret_access_key = authz.minio_secret_key, config = boto_config)
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = authz.max_workers,
        )
        def create_s3t(self):
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        
        cls.s3t = create_s3t
        #cls.s3t = s3transfer.create_transfer_manager(cls.boto, transfer_config)

    #@classmethod
    def build_filesystems(cls, force: bool = False, **auth_config):
        """
        Lazily inits the filesystems
        """
        if cls.fs and cls.fsa and not force: return
        if cls.fs_name == 's3fs':
            cls.build_s3fs(**auth_config)
        elif cls.fs_name == 'minio':
            cls.build_minio(**auth_config)
        elif cls.fs_name == 'gcsfs':
            cls.build_gcsfs(**auth_config)


    @classmethod
    def reload_filesystem(cls):
        """ 
        Reinitializes the Filesystem
        """
        raise NotImplementedError


def _dummy_func(*args, **kwargs) -> Optional[Any]:
    pass

async def dummy_async_func(*args, **kwargs)  -> Optional[Any]:
    pass

def create_method_fs(cfs: Type[CloudFileSystemType], name: Union[str, List[str]],  func: Optional[Callable] = None, fs_type: str = 'fs') -> Optional[Callable]:
    if not hasattr(cfs, fs_type):
        #print(f'{cfs.__name__} has no {fs_type}')
        return _dummy_func
    fs_module = getattr(cfs, fs_type)
    if not isinstance(name, list): name = [name]
    for n in name:
        if hasattr(fs_module, n):
            #print(f'{cfs.__name__}:{fs_module} has func {fs_type}:{n}')
            if func: return func(getattr(fs_module, n))
            return getattr(fs_module, n)
    #print(f'{cfs.__name__} has no func {fs_type}:{name}')
    return _dummy_func

def create_async_method_fs(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], func: Optional[Callable] = None, fs_type: str = 'fsa') -> Optional[Union[Callable, Coroutine]]:
    if not hasattr(cfs, fs_type):
        return dummy_async_func
    fs_module = getattr(cfs, fs_type)
    if not isinstance(name, list): name = [name]
    for n in name:
        if hasattr(fs_module, n):
            if func: return func(getattr(fs_module, n))
            return getattr(fs_module, n)
    return dummy_async_func

def create_staticmethod(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], fs_type: str = 'fs'):
    return create_method_fs(cfs, name = name, func = staticmethod, fs_type = fs_type)


def create_async_coro(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], fs_type: str = 'fs'):
    return create_async_method_fs(cfs, name = name, func = func_as_method_coro, fs_type = fs_type)


class BaseAccessor(NormalAccessor):
    """Dummy Accessor class
    """
    class CloudFileSystem(metaclass=CloudFileSystemType):
        pass
    
    info: Callable = create_staticmethod(CloudFileSystem, 'info')
    stat: Callable = create_staticmethod(CloudFileSystem, 'stat')
    size: Callable = create_staticmethod(CloudFileSystem, 'size')
    exists: Callable = create_staticmethod(CloudFileSystem, 'exists')
    is_dir: Callable = create_staticmethod(CloudFileSystem, 'isdir')
    is_file: Callable = create_staticmethod(CloudFileSystem, 'isfile')
    copy: Callable = create_staticmethod(CloudFileSystem, 'copy')
    copy_file: Callable = create_staticmethod(CloudFileSystem, 'cp_file')
    get_file: Callable = create_staticmethod(CloudFileSystem, 'get_file')
    put_file: Callable = create_staticmethod(CloudFileSystem, 'put_file')
    metadata: Callable = create_staticmethod(CloudFileSystem, ['metadata', 'info'])
    checksum: Callable = create_staticmethod(CloudFileSystem, 'checksum')

    get: Callable = create_staticmethod(CloudFileSystem, 'get')
    put: Callable = create_staticmethod(CloudFileSystem, 'put')

    open: Callable = create_method_fs(CloudFileSystem, 'open')
    listdir: Callable = create_method_fs(CloudFileSystem, 'ls')    
    glob: Callable = create_method_fs(CloudFileSystem, 'glob')
    find: Callable = create_method_fs(CloudFileSystem, 'find')
    touch: Callable = create_method_fs(CloudFileSystem, 'touch')
    cat: Callable = create_method_fs(CloudFileSystem, 'cat')
    cat_file: Callable = create_method_fs(CloudFileSystem, 'cat_file')
    
    
    
    pipe: Callable = create_method_fs(CloudFileSystem, 'pipe')
    pipe_file: Callable = create_method_fs(CloudFileSystem, 'pipe_file')
    
    mkdir: Callable = create_method_fs(CloudFileSystem, 'mkdir')
    makedirs: Callable = create_method_fs(CloudFileSystem, ['makedirs', 'mkdirs'])
    unlink: Callable = create_method_fs(CloudFileSystem, 'rm_file')
    rmdir: Callable = create_method_fs(CloudFileSystem, 'rmdir')
    rename : Callable = create_method_fs(CloudFileSystem, 'rename')
    replace : Callable = create_method_fs(CloudFileSystem, 'rename')
    remove : Callable = create_method_fs(CloudFileSystem, 'rm')
    rm : Callable = create_staticmethod(CloudFileSystem, 'rm')
    rm_file : Callable = create_staticmethod(CloudFileSystem, 'rm_file')
    
    modified: Callable = create_method_fs(CloudFileSystem, 'modified')
    url: Callable = create_method_fs(CloudFileSystem, 'url')
    ukey: Callable = create_method_fs(CloudFileSystem, 'ukey')
    setxattr: Callable = create_method_fs(CloudFileSystem, 'setxattr')
    invalidate_cache: Callable = create_method_fs(CloudFileSystem, 'invalidate_cache')
    
    filesys: ClassVar = CloudFileSystem.fs
    async_filesys: ClassVar = CloudFileSystem.fsa

    boto: ClassVar = CloudFileSystem.boto
    s3t: Callable = CloudFileSystem.s3t
    
    # Async Methods
    async_stat: Callable = create_async_coro(CloudFileSystem, 'stat')
    async_touch: Callable = create_async_coro(CloudFileSystem, 'touch')
    async_ukey: Callable = create_async_coro(CloudFileSystem, 'ukey')
    async_size: Callable = create_async_coro(CloudFileSystem, 'size')
    async_url: Callable = create_async_coro(CloudFileSystem, 'url')
    async_setxattr: Callable = create_async_coro(CloudFileSystem, 'setxattr')
    async_modified: Callable = create_async_coro(CloudFileSystem, 'modified')
    async_invalidate_cache: Callable = create_async_coro(CloudFileSystem, 'invalidate_cache')
    async_rename: Callable = create_async_coro(CloudFileSystem, 'rename')
    async_replace: Callable = create_async_coro(CloudFileSystem, 'rename')

    async_info: Callable = create_async_method_fs(CloudFileSystem, 'async_info')
    async_exists: Callable = create_async_method_fs(CloudFileSystem, 'async_exists')
    async_glob: Callable = create_async_method_fs(CloudFileSystem, 'async_glob')
    async_find: Callable = create_async_method_fs(CloudFileSystem, 'async_find')
    async_is_dir: Callable = create_async_method_fs(CloudFileSystem, 'async_isdir')
    async_is_file: Callable = create_async_method_fs(CloudFileSystem, 'async_is_file')
    async_copy: Callable = create_async_method_fs(CloudFileSystem, 'async_copy')
    async_copy_file: Callable = create_async_method_fs(CloudFileSystem, 'async_cp_file')

    async_pipe: Callable = create_async_method_fs(CloudFileSystem, 'async_pipe')
    async_pipe_file: Callable = create_async_method_fs(CloudFileSystem, 'async_pipe_file')

    async_get: Callable = create_async_coro(CloudFileSystem, 'async_get')
    async_get_file: Callable = create_async_coro(CloudFileSystem, 'async_get_file')
    
    async_put: Callable = create_async_method_fs(CloudFileSystem, 'async_put')
    async_put_file: Callable = create_async_method_fs(CloudFileSystem, 'async_put_file')
    async_metadata: Callable = create_async_method_fs(CloudFileSystem, 'async_info')
    async_open: Callable = create_async_method_fs(CloudFileSystem, '_open')
    async_mkdir: Callable = create_async_method_fs(CloudFileSystem, 'async_mkdir')
    async_makedirs: Callable = create_async_method_fs(CloudFileSystem, 'async_makedirs')
    async_unlink: Callable = create_async_method_fs(CloudFileSystem, 'async_rm_file')
    async_rmdir: Callable = create_async_method_fs(CloudFileSystem, 'async_rmdir')
    async_remove: Callable = create_async_method_fs(CloudFileSystem, 'async_rm')
    async_rm: Callable = create_async_method_fs(CloudFileSystem, 'async_rm')
    async_rm_file: Callable = create_async_coro(CloudFileSystem, 'async_rm_file')
    async_listdir: Callable = create_async_method_fs(CloudFileSystem, ['async_listdir', 'async_list_objects'])

    @classmethod
    def reload_cfs(cls, **kwargs):
        cls.CloudFileSystem.build_filesystems(**kwargs)
        cls.info: Callable = create_staticmethod(cls.CloudFileSystem, 'info')
        cls.stat: Callable = create_staticmethod(cls.CloudFileSystem, 'stat')
        cls.size: Callable = create_staticmethod(cls.CloudFileSystem, 'size')
        cls.exists: Callable = create_staticmethod(cls.CloudFileSystem, 'exists')
        cls.is_dir: Callable = create_staticmethod(cls.CloudFileSystem, 'isdir')
        cls.is_file: Callable = create_staticmethod(cls.CloudFileSystem, 'isfile')
        cls.copy: Callable = create_staticmethod(cls.CloudFileSystem, 'copy')
        cls.copy_file: Callable = create_staticmethod(cls.CloudFileSystem, 'cp_file')
        cls.get_file: Callable = create_staticmethod(cls.CloudFileSystem, 'get_file')
        cls.put_file: Callable = create_staticmethod(cls.CloudFileSystem, 'put_file')
        cls.metadata: Callable = create_staticmethod(cls.CloudFileSystem, ['metadata', 'info'])

        cls.open: Callable = create_method_fs(cls.CloudFileSystem, 'open')
        cls.listdir: Callable = create_method_fs(cls.CloudFileSystem, 'ls')    
        cls.glob: Callable = create_staticmethod(cls.CloudFileSystem, 'glob')
        cls.get: Callable = create_staticmethod(cls.CloudFileSystem, 'get')
        cls.put: Callable = create_staticmethod(cls.CloudFileSystem, 'put')
        
        cls.checksum: Callable = create_method_fs(cls.CloudFileSystem, 'checksum')
        cls.cat: Callable = create_staticmethod(cls.CloudFileSystem, 'cat')
        cls.cat_file: Callable = create_staticmethod(cls.CloudFileSystem, 'cat_file')
        
        cls.pipe: Callable = create_staticmethod(cls.CloudFileSystem, 'pipe')
        cls.pipe_file: Callable = create_staticmethod(cls.CloudFileSystem, 'pipe_file')
    

        cls.find: Callable = create_method_fs(cls.CloudFileSystem, 'find')
        cls.touch: Callable = create_method_fs(cls.CloudFileSystem, 'touch')
        
        
        cls.mkdir: Callable = create_method_fs(cls.CloudFileSystem, 'mkdir')
        cls.makedirs: Callable = create_method_fs(cls.CloudFileSystem, ['makedirs', 'mkdirs'])
        cls.unlink: Callable = create_method_fs(cls.CloudFileSystem, 'rm_file')
        cls.rmdir: Callable = create_method_fs(cls.CloudFileSystem, 'rmdir')
        cls.rename : Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.replace : Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.rm : Callable = create_staticmethod(cls.CloudFileSystem, 'rm')
        cls.rm_file : Callable = create_staticmethod(cls.CloudFileSystem, 'rm_file')
        
        cls.remove : Callable = create_method_fs(cls.CloudFileSystem, 'rm')
        cls.modified: Callable = create_method_fs(cls.CloudFileSystem, 'modified')
        cls.setxattr: Callable = create_method_fs(cls.CloudFileSystem, 'setxattr')
        cls.url: Callable = create_method_fs(cls.CloudFileSystem, 'url')
        cls.ukey: Callable = create_method_fs(cls.CloudFileSystem, 'ukey')
        cls.invalidate_cache: Callable = create_method_fs(cls.CloudFileSystem, 'invalidate_cache')
        
        cls.filesys = cls.CloudFileSystem.fs
        cls.async_filesys = cls.CloudFileSystem.fsa
        cls.boto = cls.CloudFileSystem.boto
        cls.s3t = cls.CloudFileSystem.s3t
        
        # Async Methods
        cls.async_stat: Callable = create_async_coro(cls.CloudFileSystem, 'stat')
        cls.async_touch: Callable = create_async_coro(cls.CloudFileSystem, 'touch')
        cls.async_ukey: Callable = create_async_coro(cls.CloudFileSystem, 'ukey')
        cls.async_size: Callable = create_async_coro(cls.CloudFileSystem, 'size')
        cls.async_url: Callable = create_async_coro(cls.CloudFileSystem, 'url')
        cls.async_setxattr: Callable = create_async_coro(cls.CloudFileSystem, 'setxattr')

        cls.async_modified: Callable = create_async_coro(cls.CloudFileSystem, 'modified')
        cls.async_invalidate_cache: Callable = create_async_coro(cls.CloudFileSystem, 'invalidate_cache')
        cls.async_rename: Callable = create_async_coro(cls.CloudFileSystem, 'rename')
        cls.async_replace: Callable = create_async_coro(cls.CloudFileSystem, 'rename')

        cls.async_info: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_info')
        cls.async_exists: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_exists')

        cls.async_glob: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_glob')
        cls.async_find: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_find')
        
        cls.async_cat: Callable = create_async_coro(cls.CloudFileSystem, 'async_cat')
        cls.async_cat_file: Callable = create_async_coro(cls.CloudFileSystem, 'async_cat_file')
        
        cls.async_pipe: Callable = create_async_coro(cls.CloudFileSystem, 'async_pipe')
        cls.async_pipe_file: Callable = create_async_coro(cls.CloudFileSystem, 'async_pipe_file')
        
        cls.async_is_dir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_isdir')
        cls.async_is_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_is_file')
        cls.async_copy: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_copy')
        cls.async_copy_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_cp_file')
        cls.async_get: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_get')
        cls.async_get_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_get_file')
        cls.async_put: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_put')
        cls.async_put_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_put_file')
        cls.async_metadata: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_info')
        cls.async_open: Callable = create_async_method_fs(cls.CloudFileSystem, '_open')
        cls.async_mkdir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_mkdir')
        cls.async_makedirs: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_makedirs')
        cls.async_unlink: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm_file')
        cls.async_rm_file: Callable = create_async_coro(cls.CloudFileSystem, 'async_rm_file')
        cls.async_rmdir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rmdir')
        cls.async_remove: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm')
        cls.async_rm: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm')
        cls.async_listdir: Callable = create_async_method_fs(cls.CloudFileSystem, ['async_listdir', 'async_list_objects'])



class GCP_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 'gcsfs.GCSFileSystem' = None
    fsa: 'gcsfs.GCSFileSystem' = None
    fs_name: str = 'gcsfs'

class AWS_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fs_name: str = 's3fs'
    boto: 'boto3.session.Session' = None
    s3t: Callable = None

class Minio_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fs_name: str = 'minio'
    boto: 'boto3.session.Session' = None
    s3t: Callable = None

class GCP_Accessor(BaseAccessor):
    """
    GCP Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(GCP_CloudFileSystem):
        pass

class AWS_Accessor(BaseAccessor):
    """
    AWS Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(AWS_CloudFileSystem):
        pass

class Minio_Accessor(BaseAccessor):
    """
    S3 Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(Minio_CloudFileSystem):
        pass

_GCPAccessor: GCP_Accessor = None
_AWSAccessor: AWS_Accessor = None
_MinioAccessor: Minio_Accessor = None

def _get_gcp_accessor(**kwargs) -> GCP_Accessor:
    global _GCPAccessor, GCP_Accessor
    if not _GCPAccessor:
        GCP_CloudFileSystem.build_filesystems(**kwargs)
        GCP_Accessor.reload_cfs(**kwargs)
        _GCPAccessor = GCP_Accessor()
    return _GCPAccessor

def _get_aws_accessor(**kwargs) -> AWS_Accessor:
    global _AWSAccessor, AWS_Accessor
    if not _AWSAccessor:
        AWS_CloudFileSystem.build_filesystems(**kwargs)
        AWS_Accessor.reload_cfs(**kwargs)
        _AWSAccessor = AWS_Accessor()
    return _AWSAccessor

def _get_minio_accessor(**kwargs) -> Minio_Accessor:
    global _MinioAccessor, Minio_Accessor
    if not _MinioAccessor:
        Minio_CloudFileSystem.build_filesystems(**kwargs)
        Minio_Accessor.reload_cfs(**kwargs)
        _MinioAccessor = Minio_Accessor()
    return _MinioAccessor

_accessor_getters = {
    'gs': _get_gcp_accessor,
    's3': _get_aws_accessor,
    'minio': _get_minio_accessor
}
_cfs_getters = {
    'gs': GCP_CloudFileSystem,
    's3': AWS_CloudFileSystem,
    'minio': Minio_CloudFileSystem
}

AccessorLike = Union[
    BaseAccessor,
    GCP_Accessor,
    AWS_Accessor,
    Minio_Accessor
]
CloudFileSystemLike = Union[
    CloudFileSystemType,
    GCP_CloudFileSystem,
    AWS_CloudFileSystem,
    Minio_CloudFileSystem
]

def get_accessor(name: str, **kwargs) -> AccessorLike:
    if not _accessor_getters.get(name, None): return BaseAccessor
    return _accessor_getters[name](**kwargs)

def get_cloud_filesystem(name: str) -> Optional[CloudFileSystemLike]:
    return _cfs_getters.get(name, None)




from __future__ import annotations

import shutil
from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO
from typing import ClassVar
from hashlib import md5

from .types import *
from .base_imports import *
from .aiopath.selectors import _make_selector
from .aiopath.scandir import EntryWrapper, scandir_async, _scandir_results
from .flavours import _pathz_windows_flavour, _pathz_posix_flavour
from .utils import get_file_info

def scandir_sync(*args, **kwargs) -> Iterable[EntryWrapper]:
    results = _scandir_results(*args, **kwargs)
    yield from results

Paths = Union[Path, PathLike, str]
close = func_to_async_func(os.close)
sync_close = os.close




def generate_checksum(p: 'FilePath'):
    with p.open('rb') as f:
        file_hash = md5()
        chunk = f.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(8192)
    return file_hash.hexdigest()

def calc_etag(inputfile: 'FilePath', partsize: int = 8388608):
    md5_digests = []
    with inputfile.open('rb') as f:
        for chunk in iter(lambda: f.read(partsize), b''):
            md5_digests.append(md5(chunk).digest())
    return md5(b''.join(md5_digests)).hexdigest() + '-' + str(len(md5_digests))


class _FileAccessor(NormalAccessor):
    # Sync methods
    stat = os.stat
    info = get_file_info

    lstat = os.lstat
    open = os.open
    listdir = os.listdir
    chmod = os.chmod

    copy = shutil.copy
    copy_file = shutil.copyfile

    # Async Methods
    async_stat = func_as_method_coro(os.stat)
    async_info = func_as_method_coro(get_file_info)

    async_lstat = func_as_method_coro(os.lstat)
    async_open = func_as_method_coro(os.open)
    async_listdir = func_as_method_coro(os.listdir)
    async_chmod = func_as_method_coro(os.chmod)

    async_copy = func_as_method_coro(shutil.copy)
    async_copy_file = func_as_method_coro(shutil.copyfile)

    if hasattr(NormalAccessor, 'lchmod'):
        lchmod = NormalAccessor.lchmod
        async_lchmod = method_as_method_coro(NormalAccessor.lchmod)

    mkdir = os.mkdir
    unlink = os.unlink

    async_mkdir = func_as_method_coro(os.mkdir)
    async_unlink = func_as_method_coro(os.unlink)

    if hasattr(NormalAccessor, 'link'):
        link = NormalAccessor.link
        async_link = method_as_method_coro(NormalAccessor.link)

    rmdir = os.rmdir
    rename = os.rename
    replace = os.replace
    symlink = staticmethod(NormalAccessor.symlink)
    utime = os.utime
    readlink = NormalAccessor.readlink
    remove = os.remove

    async_rmdir = func_as_method_coro(os.rmdir)
    async_rename = func_as_method_coro(os.rename)
    async_replace = func_as_method_coro(os.replace)
    async_symlink = staticmethod(method_as_method_coro(NormalAccessor.symlink))
    async_utime = func_as_method_coro(os.utime)
    async_readlink = method_as_method_coro(NormalAccessor.readlink)
    async_remove = func_as_method_coro(os.remove)


    def owner(self, path: str) -> str:
        try:
            import pwd
            stat = self.stat(path)
            return pwd.getpwuid(stat.st_uid).pw_name

        except ImportError:
            raise NotImplementedError("Path.owner() is unsupported on this system")

    async def async_owner(self, path: str) -> str:
        try:
            import pwd
            stat = await self.async_stat(path)
            return pwd.getpwuid(stat.st_uid).pw_name
        except ImportError: raise NotImplementedError("Path.owner() is unsupported on this system")

    def group(self, path: str) -> str:
        try:
            import grp
            stat = self.stat(path)
            return grp.getgrgid(stat.st_gid).gr_name

        except ImportError: raise NotImplementedError("Path.group() is unsupported on this system")

    async def async_group(self, path: str) -> str:
        try:
            import grp
            stat = await self.async_stat(path)
            return grp.getgrgid(stat.st_gid).gr_name

        except ImportError: raise NotImplementedError("Path.group() is unsupported on this system")

    def scandir(self, *args, **kwargs) -> Iterable[EntryWrapper]:
        yield from scandir_sync(*args, **kwargs)

    async def async_scandir(self, *args, **kwargs) -> AsyncIterable[EntryWrapper]:
        async for entry in scandir_async(*args, **kwargs):
            yield entry

_pathz_accessor = _FileAccessor()


class FilePurePath(PurePath):
    _prefix: str = None
    _provider: str = None
    _win_pathz: ClassVar = 'PureFileWindowsPath'
    _posix_pathz: ClassVar = 'PureFilePosixPath'

    def _init(self, template: Optional[PurePath] = None):
        self._accessor = _pathz_accessor

    def __new__(cls, *args):
        if cls is FilePurePath or issubclass(cls, FilePurePath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        return cls._from_parts(args)


class PureFilePosixPath(FilePurePath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    __slots__ = ()


class PureFileWindowsPath(FilePurePath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    __slots__ = ()


class FilePath(Path, FilePurePath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor = _pathz_accessor
    _prefix = None
    _provider = None
    _win_pathz: ClassVar = 'FileWindowsPath'
    _posix_pathz: ClassVar = 'FilePosixPath'

    def _init(self, template: Optional[FilePath] = None):
        self._accessor = _pathz_accessor
        self._closed = False
        self._fileio = None

    def __new__(cls, *args, **kwargs):
        #if cls is FilePath: cls = FileWindowsPath if os.name == 'nt' else FilePosixPath
        if cls is FilePath or issubclass(cls, FilePath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(args, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self

    @property
    def _path(self) -> str:
        return self._cloudstr if self.is_cloud else str(self)
    
    @property
    def checksum(self):
        return generate_checksum(self)
    
    @property
    def etag(self):
        return calc_etag(self)

    @property
    def _cloudpath(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return None
    
    @property
    def _bucket(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return None
    
    @property
    def _bucketstr(self) -> Optional[str]:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return None
    
    @property
    def _pathkeys(self) -> Optional[str]:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return None
    
    @property
    def _cloudstr(self) -> Optional[str]:
        """
        Reconstructs the proper cloud URI
        """
        return None
    
    @property
    def posix_(self):
        """Return the string representation of the path with forward (/)
        slashes."""
        f = self._flavour
        return str(self).replace(f.sep, '/')

    @property
    def string(self) -> str:
        return self._cloudstr if self.is_cloud else self.posix_

    
    @property
    def filename_(self) -> str:
        """
        Returns the filename if is file, else ''
        """
        if self.is_file(): return self.parts[-1]
        return ''

    @property
    def ext_(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def extension(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix
    
    @property
    def is_cloud(self) -> bool:
        return False
    
    @property
    def is_pathz(self) -> bool:
        return True

    @property
    def exists_(self) -> bool:
        return self.exists()
    
    @property
    def is_file_(self) -> bool:
        return self.is_file()
    
    @property
    def is_dir_(self) -> bool:
        return self.is_dir()
    
    @property
    def home_(self) -> FilePath:
        return self.home()

    @property
    async def async_exists_(self) -> bool:
        return await self.async_exists()
    
    @property
    async def async_is_file_(self) -> bool:
        return await self.async_is_file()
    
    @property
    async def async_is_dir_(self) -> bool:
        return await self.async_is_dir()
    
    @property
    async def async_home_(self) -> FilePath:
        return await self.async_home()
    
    @staticmethod
    def _get_pathlike(path: PathLike):
        """
        Returns the path of the file.
        """
        from lazy.io.pathz_v2 import get_path
        return get_path(path)
    
    def open(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if self._closed: self._raise_closed()
        if 'b' in mode:
            return io.open(self, mode = mode, buffering = buffering, opener=self._opener)
        return io.open(self, mode, buffering, encoding, errors, newline, opener=self._opener)

    
    def async_open(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if 'b' in mode:
            return get_handle(self._path, mode = mode, buffering = buffering)
        return get_handle(self._path, mode, encoding=encoding, errors=errors, newline=newline)
    
    def reader(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if self._closed: self._raise_closed()
        return io.open(self, mode, buffering, encoding, errors, newline, opener=self._opener)
    
    def async_reader(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_handle(self._path, mode, encoding=encoding, errors=errors, newline=newline)
    
    def appender(self, mode: FileMode = 'a', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if self._closed: self._raise_closed()
        return io.open(self, mode, buffering, encoding, errors, newline, opener=self._opener)
    
    def async_appender(self, mode: FileMode = 'a', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_handle(self._path, mode, encoding=encoding, errors=errors, newline=newline)
    
    def writer(self, mode: FileMode = 'w', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        if self._closed: self._raise_closed()
        return io.open(self, mode, buffering, encoding, errors, newline, opener=self._opener)
    
    def async_writer(self, mode: FileMode = 'w', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_handle(self._path, mode, encoding=encoding, errors=errors, newline=newline)

    def read_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS) -> str:
        with self.open('r', encoding=encoding, errors=errors) as file:
            return file.read()

    async def async_read_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS) -> str:
        async with self.async_open('r', encoding=encoding, errors=errors) as file:
            return await file.read()

    def read_bytes(self) -> bytes:
        with self.open('rb') as file:
            return file.read()

    async def async_read_bytes(self) -> bytes:
        async with self.async_open('rb') as file:
            return await file.read()

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode='wb') as f:
            return f.write(data)

    async def async_write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        async with self.async_open(mode='wb') as f:
            return await f.write(data)

    def append_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        with self.open(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = f.write(data)
            n += f.write(newline)
            return n

    async def async_append_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.async_open(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = await f.write(data)
            n += await f.write(newline)
            return n

    def write_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        with self.open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    async def async_write_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.async_open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return await f.write(data)

    def readlink(self) -> FilePath:
        """
        Return the path to which the symbolic link points.
        """
        path: str = self._accessor.readlink(self)
        obj = self._from_parts((path,), init=False)
        obj._init(template=self)
        return obj

    async def async_readlink(self) -> FilePath:
        """
        Return the path to which the symbolic link points.
        """
        path: str = await self._accessor.async_readlink(self)
        obj = self._from_parts((path,), init=False)
        obj._init(template=self)
        return obj

    def _opener(self, name, flags, mode=0o666):
        # A stub for the opener argument to built-in open()
        return self._accessor.open(self, flags, mode)
    
    def _async_opener(self, name, flags, mode=0o666):
        # A stub for the opener argument to built-in open()
        return self._accessor.async_open(self, flags, mode)

    def _raw_open(self, flags: int, mode: int = 0o777) -> int:
        """
        Open the file pointed by this path and return a file descriptor,
        as os.open() does.
        """
        if self._closed: self._raise_closed()
        return self._accessor.open(self, flags, mode)

    async def _async_raw_open(self, flags: int, mode: int = 0o777) -> int:
        """
        Open the file pointed by this path and return a file descriptor,
        as os.open() does.
        """
        return await self._accessor.async_open(self, flags, mode)

    def touch(self, mode: int = 0o666, exist_ok: bool = True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: self._accessor.utime(self, None)
            # Avoid exception chaining
            except OSError: pass
            else: return

        flags: int = os.O_CREAT | os.O_WRONLY
        if not exist_ok: flags |= os.O_EXCL
        fd = self._raw_open(flags, mode)
        sync_close(fd)

    async def async_touch(self, mode: int = 0o666, exist_ok: bool = True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: await self._accessor.async_utime(self, None)
            # Avoid exception chaining
            except OSError: pass
            else: return

        flags: int = os.O_CREAT | os.O_WRONLY
        if not exist_ok: flags |= os.O_EXCL
        fd = await self._async_raw_open(flags, mode)
        await close(fd)

    def mkdir(self, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """
        Create a new directory at this given path.
        """
        try: self._accessor.mkdir(self, mode)

        except FileNotFoundError:
            if not parents or self.parent == self: raise
            self.parent.mkdir(parents=True, exist_ok=True)
            self.mkdir(mode, parents=False, exist_ok=exist_ok)

        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not self.is_dir(): raise

    async def async_mkdir(self, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """
        Create a new directory at this given path.
        """
        try: await self._accessor.async_mkdir(self, mode)

        except FileNotFoundError:
            if not parents or self.parent == self: raise
            await self.parent.async_mkdir(parents=True, exist_ok=True)
            await self.async_mkdir(mode, parents=False, exist_ok=exist_ok)

        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not await self.async_is_dir(): raise

    async def chmod(self, mode: int):
        """
        Change the permissions of the path, like os.chmod().
        """
        self._accessor.chmod(self, mode)

    async def async_chmod(self, mode: int):
        """
        Change the permissions of the path, like os.chmod().
        """
        await self._accessor.async_chmod(self, mode)

    def lchmod(self, mode: int):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        self._accessor.lchmod(self, mode)

    async def async_lchmod(self, mode: int):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        await self._accessor.async_lchmod(self, mode)

    def unlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: self._accessor.unlink(self)
        except FileNotFoundError:
            if not missing_ok: raise

    async def async_unlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: await self._accessor.async_unlink(self)
        except FileNotFoundError:
            if not missing_ok: raise

    def copy(self, dest: PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copies the File to the Dir/File.
        """
        dest: 'FilePath' = self._get_pathlike(dest)
        if dest.is_dir() and self.is_file():
            dest = dest.joinpath(self.filename_)
        
        if dest.exists() and not overwrite and dest.is_file():
            if skip_errors: return dest
            raise Exception(f'File {dest._path} exists')

        if not dest.is_cloud:
            self._accessor.copy(self._path, dest._path, **kwargs)
            return dest
        dest._accessor.put(self._path, dest._path, recursive)
        return dest
    
    async def async_copy(self, dest: PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        dest: 'FilePath' = self._get_pathlike(dest)
        if await dest.async_is_dir() and self.async_is_file():
            dest = dest.joinpath(self.filename_)
        
        if await dest.async_exists() and not overwrite and await dest.async_is_file():
            if skip_errors: return dest
            raise Exception(f'File {dest._path} exists')

        if not dest.is_cloud:
            await self._accessor.async_copy(self._path, dest._path, **kwargs)
            return dest
        await dest._accessor.async_put(self._path, dest._path, recursive)
        return dest
    
    def copy_file(self, dest: PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        dest: 'FilePath' = self._get_pathlike(dest)
        if dest.is_dir() and self.is_file():
            dest = dest.joinpath(self.filename_)
        
        if dest.exists() and not overwrite and dest.is_file():
            if skip_errors: return dest
            raise Exception(f'File {dest._path} exists')

        if not dest.is_cloud:
            self._accessor.copy_file(self._path, dest._path, **kwargs)
            return dest
        dest._accessor.put_file(self._path, dest._path, recursive)
        return dest

    async def async_copy_file(self, dest: PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        dest: 'FilePath' = self._get_pathlike(dest)
        if await dest.async_is_dir() and self.async_is_file():
            dest = dest.joinpath(self.filename_)
        
        if await dest.async_exists() and not overwrite and await dest.async_is_file():
            if skip_errors: return dest
            raise Exception(f'File {dest._path} exists')

        if not dest.is_cloud:
            await self._accessor.async_copy_file(self._path, dest._path, **kwargs)
            return dest
        await dest._accessor.async_put_file(self._path, dest._path, recursive)
        return dest

    def rm(self, **kwargs):
        """
        Remove this file or dir
        """
        if self.is_dir: return self.rmdir(**kwargs)
        
        return self._accessor.remove(self)
    
    async def async_rm(self, **kwargs):
        """
        Remove this file or dir
        """
        if self.is_dir: return await self.async_rmdir(**kwargs)
        await self._accessor.async_remove(self)

    def rm_file(self, **kwargs):
        """
        Remove this file 
        """
        
        self._accessor.remove(self)
    
    async def async_rm_file(self, **kwargs):
        """
        Remove this file 
        """
        
        return await self._accessor.async_remove(self)

    def rmdir(self, force: bool = False, recursive: bool = True, skip_errors: bool = True):
        """
        Remove this directory.  The directory must be empty.
        """
        
        self._accessor.rmdir(self)

    async def async_rmdir(self):
        """
        Remove this directory.  The directory must be empty.
        """
        await self._accessor.async_rmdir(self)

    def cat(self, as_bytes: bool = False, **kwargs):
        """
        Fetch paths’ contents
        """
        return self.read_bytes() if as_bytes else self.read_text()
    
    async def async_cat(self, as_bytes: bool = False, **kwargs):
        """
        Fetch paths’ contents
        """
        if as_bytes: return await self.async_read_bytes()
        return await self.async_read_text()
    
    def cat_file(self, as_bytes: bool = False, **kwargs):
        """
        """
        return self.cat(as_bytes, **kwargs)
    
    async def async_cat_file(self, as_bytes: bool = False, **kwargs):
        """
        Parameters
        start, end: int
            Bytes limits of the read. If negative, backwards from end, like usual python slices. Either can be None for start or end of file, respectively

        kwargs: passed to ``open()``.
        """
        return await self.async_cat(as_bytes, **kwargs)


    def pipe(self, value: Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return self.write_bytes(value, **kwargs)

    async def async_pipe(self, value: Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return await self.async_write_bytes(value, **kwargs)

    def pipe_file(self, value: Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return self.write_bytes(value, **kwargs)

    async def async_pipe_file(self, value: Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return await self.async_write_bytes(value, **kwargs)

    def link_to(self, target: str):
        """
        Create a hard link pointing to a path named target.
        """
        self._accessor.link_to(self, target)
    
    async def async_link_to(self, target: str):
        """
        Create a hard link pointing to a path named target.
        """
        await self._accessor.async_link_to(self, target)

    def rename(self, target: Union[str, FilePath]) -> FilePath:
        """
        Rename this path to the target path.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        self._accessor.rename(self, target)
        return type(self)(target)
    
    async def async_rename(self, target: Union[str, FilePath]) -> FilePath:
        """
        Rename this path to the target path.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        await self._accessor.async_rename(self, target)
        return type(self)(target)

    def replace(self, target: str) -> FilePath:
        """
        Rename this path to the target path, overwriting if that path exists.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        self._accessor.replace(self, target)
        return type(self)(target)
    
    async def async_replace(self, target: str) -> FilePath:
        """
        Rename this path to the target path, overwriting if that path exists.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        await self._accessor.async_replace(self, target)
        return type(self)(target)

    def symlink_to(self, target: str, target_is_directory: bool = False):
        """
        Make this path a symlink pointing to the given path.
        Note the order of arguments (self, target) is the reverse of os.symlink's.
        """
        self._accessor.symlink(target, self, target_is_directory)
    
    async def async_symlink_to(self, target: str, target_is_directory: bool = False):
        """
        Make this path a symlink pointing to the given path.
        Note the order of arguments (self, target) is the reverse of os.symlink's.
        """
        await self._accessor.async_symlink(target, self, target_is_directory)

    def exists(self) -> bool:
        """
        Whether this path exists.
        """
        try: self.stat()
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False
        return True

    async def async_exists(self) -> bool:
        """
        Whether this path exists.
        """
        try: await self.async_stat()
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False
        return True

    @classmethod
    def cwd(cls: type) -> str:
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        cwd: str = os.getcwd()
        return cls(cwd)

    @classmethod
    def home(cls: type) -> FilePath:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        homedir: str = cls()._flavour.gethomedir(None)
        return cls(homedir)

    @classmethod
    async def async_home(cls: type) -> FilePath:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        coro = cls()._flavour.async_gethomedir(None)
        homedir: str = await coro
        return cls(homedir)

    def samefile(self, other_path: Union[FilePath, Paths]) -> bool:
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        if isinstance(other_path, Paths.__args__): other_path = FilePath(other_path)
        if isinstance(other_path, FilePath):
            try: other_st = other_path.stat()
            except AttributeError: other_st = self._accessor.stat(other_path)

        else:
            try: other_st = other_path.stat()
            except AttributeError: other_st = other_path._accessor.stat(other_path)
        return os.path.samestat(self.stat(), other_st)

    async def async_samefile(self, other_path: Union[FilePath, Paths]) -> bool:
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        if isinstance(other_path, Paths.__args__): other_path = FilePath(other_path)
        if isinstance(other_path, FilePath):
            try: other_st = await other_path.async_stat()
            except AttributeError: other_st = await self._accessor.async_stat(other_path)

        else:
            try: other_st = await to_thread(other_path.stat)
            except AttributeError: other_st = await to_thread(other_path._accessor.stat, other_path)

        return os.path.samestat(await self.async_stat(),other_st)

    def iterdir(self) -> Iterable[FilePath]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for name in self._accessor.listdir(self):
            if name in {'.', '..'}: continue
            yield self._make_child_relpath(name)

    async def async_iterdir(self) -> AsyncIterable[FilePath]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for name in await self._accessor.async_listdir(self):
            if name in {'.', '..'}: continue
            yield self._make_child_relpath(name)

    def glob(self, pattern: str) -> Iterable[FilePath]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))

        drv, root, pattern_parts = self._flavour.parse_parts((pattern,))
        if drv or root: raise NotImplementedError("Non-relative patterns are unsupported")
        selector = _sync_make_selector(tuple(pattern_parts), self._flavour)
        yield from selector.select_from(self)

    async def async_glob(self, pattern: str) -> AsyncIterable[FilePath]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))

        drv, root, pattern_parts = self._flavour.parse_parts((pattern,))
        if drv or root: raise NotImplementedError("Non-relative patterns are unsupported")
        selector = _make_selector(tuple(pattern_parts), self._flavour)
        async for p in selector.select_from(self):
            yield p

    def rglob(self, pattern: str) -> Iterable[FilePath]:
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        drv, root, pattern_parts = self._flavour.parse_parts((pattern,))

        if drv or root: raise NotImplementedError("Non-relative patterns are unsupported")
        parts = ("**", *pattern_parts)
        selector = _sync_make_selector(parts, self._flavour)
        yield from selector.select_from(self)

    async def async_rglob(self, pattern: str) -> AsyncIterable[FilePath]:
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        drv, root, pattern_parts = self._flavour.parse_parts((pattern,))
        if drv or root: raise NotImplementedError("Non-relative patterns are unsupported")
        parts = ("**", *pattern_parts)
        selector = _make_selector(parts, self._flavour)
        async for p in selector.select_from(self):
            yield p

    def absolute(self) -> FilePath:
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.
        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        if self.is_absolute(): return self
        obj = self._from_parts([os.getcwd()] + self._parts, init=False)
        obj._init(template=self)
        return obj

    def resolve(self, strict: bool = False) -> FilePath:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        s: Optional[str] = self._flavour.resolve(self, strict=strict)

        if s is None:
            self.stat()
            path = self.absolute()
            s = str(path)

        # Now we have no symlinks in the path, it's safe to normalize it.
        normed: str = self._flavour.pathmod.normpath(s)
        obj = self._from_parts((normed,), init=False)
        obj._init(template=self)
        return obj

    async def async_resolve(self, strict: bool = False) -> FilePath:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        s: Optional[str] = await self._flavour.async_resolve(self, strict=strict)

        if s is None:
            await self.async_stat()
            path = await self.absolute()
            s = str(path)

        # Now we have no symlinks in the path, it's safe to normalize it.
        normed: str = self._flavour.pathmod.normpath(s)
        obj = self._from_parts((normed,), init=False)
        obj._init(template=self)
        return obj

    def stat(self) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return self._accessor.stat(self)
    
    async def async_stat(self) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return await self._accessor.async_stat(self)
    
    def info(self) -> Dict[str, Any]:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return self._accessor.info(self)
    
    async def async_info(self) -> Dict[str, Any]:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return await self._accessor.async_info(self)

    def lstat(self) -> stat_result:
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        return self._accessor.lstat(self)
    
    async def async_lstat(self) -> stat_result:
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        return await self._accessor.async_lstat(self)

    def owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        return self._accessor.owner(self)
    
    async def async_owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        return await self._accessor.async_owner(self)

    def group(self) -> str:
        """
        Return the group name of the file gid.
        """
        return self._accessor.group(self)
    
    async def async_group(self) -> str:
        """
        Return the group name of the file gid.
        """
        return await self._accessor.async_group(self)

    def is_dir(self) -> bool:
        """
        Whether this path is a directory.
        """
        try:
            stat = self.stat()
            return S_ISDIR(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False
    
    async def async_is_dir(self) -> bool:
        """
        Whether this path is a directory.
        """
        try:
            stat = await self.async_stat()
            return S_ISDIR(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    def is_symlink(self) -> bool:
        """
        Whether this path is a symbolic link.
        """
        try:
            lstat = self.lstat()
            return S_ISLNK(lstat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False
    
    async def async_is_symlink(self) -> bool:
        """
        Whether this path is a symbolic link.
        """
        try:
            lstat = await self.async_lstat()
            return S_ISLNK(lstat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    def is_file(self) -> bool:
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        try:
            stat = self.stat()
            return S_ISREG(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    async def async_is_file(self) -> bool:
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        try:
            stat = await self.async_stat()
            return S_ISREG(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    def is_mount(self) -> bool:
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        if not self.exists() or not self.is_dir(): return False
        
        try:
            parent_stat = self.parent.stat()
            parent_dev = parent_stat.st_dev
        except OSError: return False

        stat = self.stat()
        dev = stat.st_dev
        if dev != parent_dev: return True
        ino = stat.st_ino
        parent_ino = parent_stat.st_ino
        return ino == parent_ino

    async def async_is_mount(self) -> bool:
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        if not await self.async_exists() or not await self.async_is_dir(): return False
        
        try:
            parent_stat = await self.parent.async_stat()
            parent_dev = parent_stat.st_dev
        except OSError: return False

        stat = await self.async_stat()
        dev = stat.st_dev
        if dev != parent_dev: return True
        ino = stat.st_ino
        parent_ino = parent_stat.st_ino
        return ino == parent_ino

    def is_block_device(self) -> bool:
        """
        Whether this path is a block device.
        """
        try:
            stat = self.stat()
            return S_ISBLK(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    async def async_is_block_device(self) -> bool:
        """
        Whether this path is a block device.
        """
        try:
            stat = await self.async_stat()
            return S_ISBLK(stat.st_mode)

        except OSError as e:
            if not _ignore_error(e): raise
            return False

        except ValueError: return False

    def is_char_device(self) -> bool:
        """
        Whether this path is a character device.
        """
        try:
            stat = self.stat()
            return S_ISCHR(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False
    
    async def async_is_char_device(self) -> bool:
        """
        Whether this path is a character device.
        """
        try:
            stat = await self.stat()
            return S_ISCHR(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    def is_fifo(self) -> bool:
        """
        Whether this path is a FIFO.
        """
        try:
            stat = self.stat()
            return S_ISFIFO(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    async def async_is_fifo(self) -> bool:
        """
        Whether this path is a FIFO.
        """
        try:
            stat = await self.async_stat()
            return S_ISFIFO(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    def is_socket(self) -> bool:
        """
        Whether this path is a socket.
        """
        try:
            stat = self.stat()
            return S_ISSOCK(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False
    
    async def async_is_socket(self) -> bool:
        """
        Whether this path is a socket.
        """
        try:
            stat = await self.async_stat()
            return S_ISSOCK(stat.st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    def expanduser(self) -> FilePath:
        """ Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        if (not self._drv and not self._root and self._parts and self._parts[0][:1] == '~'):
            homedir = self._flavour.gethomedir(self._parts[0][1:])
            return self._from_parts([homedir] + self._parts[1:])
        return self
    
    async def async_expanduser(self) -> FilePath:
        """ Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        if (not self._drv and not self._root and self._parts and self._parts[0][:1] == '~'):
            homedir = await self._flavour.async_gethomedir(self._parts[0][1:])
            return self._from_parts([homedir] + self._parts[1:])
        return self

    def iterdir(self) -> Iterable[FilePath]:
        names = self._accessor.listdir(self)
        for name in names:
            if name in {'.', '..'}: continue
        yield self._make_child_relpath(name)
    
    async def async_iterdir(self) -> AsyncIterable[FilePath]:
        names = await self._accessor.async_listdir(self)
        for name in names:
            if name in {'.', '..'}: continue
        yield self._make_child_relpath(name)

    def _raise_closed(self):
        raise ValueError("I/O operation on closed path")
    
    def _raise_open(self):
        raise ValueError("I/O operation on already open path")
    
    # We sort of assume that it will be used to open a file
    def __enter__(self):
        #if self._fileio: self._raise_open()
        #if not self._fileio:
        #    self._fileio = self.open()
        if self._closed: self._raise_closed()
        return self

    def __exit__(self, t, v, tb):
        self._closed = True
    
    async def __aenter__(self):
        if self._closed: self._raise_closed()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._closed = True


class FilePosixPath(PosixPath, FilePath, PureFilePosixPath):
    __slots__ = ()


class FileWindowsPath(WindowsPath, FilePath, PureFileWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FilePath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FilePath.async_is_mount() is unsupported on this system")


os.PathLike.register(FilePurePath)
os.PathLike.register(FilePath)
os.PathLike.register(PureFilePosixPath)
os.PathLike.register(FileWindowsPath)
os.PathLike.register(FilePosixPath)
os.PathLike.register(PureFileWindowsPath)


__all__ = (
    'FilePurePath',
    'FilePath',
    'PureFilePosixPath',
    'FileWindowsPath',
    'FilePosixPath',
    'PureFileWindowsPath',
)



from typing import AsyncIterable, Iterable, Any, List
from os import scandir, DirEntry, stat_result

from .wrap import to_thread


class EntryWrapper:
    __slots__ = "entry",

    def __init__(self, entry: DirEntry):
        self.entry = entry

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.entry, attr)

    def __repr__(self) -> str:
        name = type(self).__name__
        return f"{name}<{self.entry}>"

    async def inode(self) -> int:
        return await to_thread(self.entry.inode)

    async def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        return await to_thread(self.entry.is_dir, follow_symlinks=follow_symlinks)

    async def is_file(self, *, follow_symlinks: bool = True) -> bool:
        return await to_thread(self.entry.is_file, follow_symlinks=follow_symlinks)

    async def is_symlink(self) -> bool:
        return await to_thread(self.entry.is_symlink)

    async def stat(self, *, follow_symlinks: bool = True) -> stat_result:
        return await to_thread(self.entry.stat, follow_symlinks=follow_symlinks)


def wrapped_scandir(*args, **kwargs) -> Iterable[EntryWrapper]:
    entries = scandir(*args, **kwargs)
    yield from map(EntryWrapper, entries)


def _scandir_results(*args, **kwargs) -> List[EntryWrapper]:
    return list(wrapped_scandir(*args, **kwargs))


async def scandir_async(*args, **kwargs) -> AsyncIterable[EntryWrapper]:
    results = await to_thread(_scandir_results, *args, **kwargs)

    for result in results:
        yield result


import os
from .baselib import pathlib
from .base import *
from .providers.gs_gcp import *
from .providers.s3_aws import *
from .providers.s3_minio import *

from typing import List, Dict, Union, Type, Tuple

FileLike = Union[
    Type[FilePurePath],
    Type[FilePath],
    Type[PureFilePosixPath],
    Type[FileWindowsPath],
    Type[FilePosixPath],
    Type[PureFileWindowsPath],
    Type[FileGSPurePath],
    Type[FileGSPath],
    Type[PureFileGSPosixPath],
    Type[FileGSWindowsPath],
    Type[FileGSPosixPath],
    Type[PureFileGSWindowsPath],
    Type[FileS3PurePath],
    Type[FileS3Path],
    Type[PureFileS3PosixPath],
    Type[FileS3WindowsPath],
    Type[FileS3PosixPath],
    Type[PureFileS3WindowsPath],
    Type[FileMinioPurePath], Type[FileMinioPath], Type[PureFileMinioPosixPath], Type[FileMinioWindowsPath], Type[FileMinioPosixPath], Type[PureFileMinioWindowsPath]

]

_PATHLIKE_CLS: Tuple[FileLike, ...] = (
    FilePurePath,
    FilePath,
    PureFilePosixPath,
    FileWindowsPath,
    FilePosixPath,
    PureFileWindowsPath,
    FileGSPurePath,
    FileGSPath,
    PureFileGSPosixPath,
    FileGSWindowsPath,
    FileGSPosixPath,
    PureFileGSWindowsPath,
    FileS3PurePath,
    FileS3Path,
    PureFileS3PosixPath,
    FileS3WindowsPath,
    FileS3PosixPath,
    PureFileS3WindowsPath,
    FileMinioPurePath, FileMinioPath, PureFileMinioPosixPath, FileMinioWindowsPath, FileMinioPosixPath, PureFileMinioWindowsPath
)

FileSysLike = Union[
    Type[AWSFileSystem],
    Type[GCPFileSystem],
    Type[MinioFileSystem],
]

PathLike = Union[str, os.PathLike, FileLike]

_PREFIXES_TO_CLS: Dict[str, FileLike] = {
    'gs://': FileGSPath,
    's3://': FileS3Path,
    #'s3a://': FileMinioPath,
    'mio://': FileMinioPath,
    'minio://': FileMinioPath,
    #'minio://': cloud.PosixMinioPath,
    #'s3compat://': cloud.PosixS3CompatPath,
}


def as_path(path: PathLike) -> FileLike:
    """
    Given a path-like object, return a path-like object
    
    Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.

    Args:
      path (PathLike): Pathlike object.
    
    Returns:
      A pathlib-like abstraction.
    """
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    
            # str is URI (e.g. `gs://`, `github://`,...)
            return _PREFIXES_TO_CLS[uri_splits[0] + '://'](path)
        return FilePath(path)
    elif isinstance(path, _PATHLIKE_CLS):
        return path
    elif isinstance(path, os.PathLike):
        return FilePath(path)
    else: raise TypeError(f'Invalid path type: {path!r}')


def get_userhome(as_pathz: bool = True):
    h = os.path.expanduser('~')
    if as_pathz: return as_path(h)
    return h

def get_cwd():
    return os.getcwd()


def resolve_relative(filepath: PathLike) -> str:
    """
    If the filepath is a relative path, convert it to an absolute path
    
    Args:
      filepath (PathLike): The path to the file you want to resolve.
    
    Returns:
      A string.
    """
    if not isinstance(filepath, str): filepath = filepath.as_posix()
    if '://' in filepath: return filepath
    if filepath.startswith('~'): filepath = filepath.replace('~', get_userhome(), 1)
    elif filepath.startswith('../'): filepath = filepath.replace('..', get_cwd(), 1)
    elif filepath.startswith('..'): filepath = filepath.replace('..', pathlib.Path(get_cwd()).parent.parent.as_posix() + '/', 1)
    elif filepath.startswith('./'): filepath = filepath.replace('.', get_cwd(), 1)
    elif filepath.startswith('.'): filepath = filepath.replace('.', pathlib.Path(get_cwd()).parent.as_posix() + '/', 1)
    return filepath

def get_path(filepath: PathLike, resolve: bool = False) -> FileLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_pathlike(filepath: PathLike, resolve: bool = False) -> FileLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath


to_path = get_path

class File:
    def __new__(cls, *args, **kwargs) -> FileLike:
        return get_path(*args, **kwargs)

__all__ = (
    'get_path', 
    'as_path', 
    'PathLike', 
    'File',
    'FilePath', 
    'FileS3Path', 
    'FileGSPath', 
    'FileLike',
)


import asyncio
import multiprocessing as mp

from concurrent import futures
from functools import wraps, partial
from typing import Callable, Any, Awaitable
from anyio.to_thread import run_sync

CoroutineResult = Awaitable[Any]
CoroutineFunction = Callable[..., CoroutineResult]
CoroutineMethod = Callable[..., CoroutineResult]

num_workers = mp.cpu_count() // 2
_pool = futures.ThreadPoolExecutor(max_workers = num_workers)

async def to_thread(func: Callable, *args, **kwargs) -> Any:
    # anyio's run_sync() doesn't support passing kwargs
    func_kwargs = partial(func, *args, **kwargs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(_pool, func_kwargs)
    #return await run_sync(func_kwargs, *args)


#async def to_thread(func: Callable, *args, **kwargs) -> Any:
#    # anyio's run_sync() doesn't support passing kwargs
#    func_kwargs = partial(func, **kwargs)
#    return await run_sync(func_kwargs, *args)


def func_to_async_func(func: Callable) -> CoroutineFunction:
    @wraps(func)
    async def new_func(*args, **kwargs) -> Any:
        return await to_thread(func, *args, **kwargs)

    return new_func


method_as_method_coro = func_to_async_func


def func_as_method_coro(func: Callable) -> CoroutineMethod:
    @wraps(func)
    async def method(self, *args, **kwargs) -> Any:
        return await to_thread(func, *args, **kwargs)
    return method


def coro_as_method_coro(coro: CoroutineFunction) -> CoroutineMethod:
    @wraps(coro)
    async def method(self, *args, **kwargs) -> Any:
        return await coro(*args, **kwargs)
    return method


__all__ = (
    'NoPrefixNames',
    'ReplacePrefixNames',
    'ConfigModelMeta',
    'ConfigModel'
)

import os

from collections import namedtuple
from inspect import currentframe
from pydantic import BaseSettings
from pydantic.env_settings import SettingsSourceCallable
from typing import List, Tuple, Dict, Any, Type, TYPE_CHECKING

try:
    from pydantic.main import ModelMetaclass
except ImportError:  # pragma: no cover
    ModelMetaclass = type(BaseSettings)


NoPrefixNames: List[str] = ['base', 'authz']
ReplacePrefixNames: List[str] = ['Cfg', 'Configz', 'Config', 'Settings']

class ConfigModelMeta(ModelMetaclass):
    def __new__(mcs, name, bases, namespace):
        Config = namespace.setdefault("Config", type("Config", (), {}))
        if not hasattr(Config, "env_prefix"):
            _prefixed = False
            for n in NoPrefixNames:
                if n in name.lower(): 
                    Config.env_prefix = ""
                    _prefixed = True
                    break
            for n in ReplacePrefixNames:
                if _prefixed: continue
                if n in name:
                    Config.env_prefix = name.replace(n, "") + "_"
                    _prefixed = True
                    break
            if not _prefixed: Config.env_prefix = name + "_"

        cls = super().__new__(mcs, name, bases, namespace)
        if not getattr(Config, "auto_init", True): return cls
        else: return cls()


class ConfigModel(BaseSettings, metaclass=ConfigModelMeta):

    class Config:        
        auto_init = False
        env_prefix = ""
        arbitrary_types_allowed = True
        extra = 'allow'

    @property
    def env_values(self): 
        return {
            self.Config.env_prefix.upper() + k.upper(): v
            for k, v in self.dict().items()
        }

    def prefixed_dict(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Similar to pydantic's `dict()`, but the `env_prefix` is included in the keys,
        so they match the original environment variable names.
        For example:
            class Database(ConfigModel):
                host = "localhost"
                username = "admin"
            assert Database.dict() == {"host": "localhost", "username": "admin"}
            assert Database.prefixed_dict() == {"DATABASE_HOST": "localhost", "DATABASE_USERNAME": "admin"}
        """
        return {
            self.Config.env_prefix.upper() + k.upper(): v
            for k, v in self.dict(*args, **kwargs).items()
        }

    def __call__(self, *args, **kwargs) -> "Config":
        """
        Instantiate this class as if you were calling the class itself.
        """
        return type(self)(*args, **kwargs)
    
    def reload(self, *args, **kwargs):
        """
        Reloads the class by reinitializing and reloading environment variables
        if they've changed.
        
        os.environ['DATABASE_HOST'] = 'localhost'
        assert Database.host == 'localhost'
        os.environ['DATABASE_HOST'] = '127.0.0.1'
        Database.reload()
        assert Database.host == '127.0.0.1'
        """
        self = self.__init__(*args, **kwargs)
    
    def update_config(self, **kwargs) -> Type["Config"]:
        if not kwargs: return
        self.reload(**kwargs)
    

def populate_globals(globs=None):
    """
    Search for instances of `ConfigModel` in the global variables in the calling context
    and then update the global variables with the `prefixed_dict()` of those `ConfigModel` isntances.
    For example, this code::
        class Database(ConfigModel):
            host = "localhost"
            username = "admin"
        populate_globals()
    will result in global variables `DATABASE_HOST` and `DATABASE_USERNAME`.
    This is useful in e.g. Django where settings need to be declared at the global level.
    You can pass your own dict for the function to use instead of the current global variables.
    """
    if globs is None:
        globs = currentframe().f_back.f_globals

    for cls in list(globs.values()):
        if isinstance(cls, ConfigModel):
            globs.update(cls.prefixed_dict())


# handle posix stat stuff
import os
import sys
#import stat
import typing
import hashlib

#from datetime import datetime
#from urllib.parse import quote
from email.utils import formatdate
from mimetypes import guess_type as mimetypes_guess_type


try:
    # check if the Python version supports the parameter
    # using usedforsecurity=False to avoid an exception on FIPS systems
    # that reject usedforsecurity=True
    hashlib.md5(b"data", usedforsecurity=False)  # type: ignore[call-arg]

    def md5_hexdigest(
        data: bytes, *, usedforsecurity: bool = True
    ) -> str:  # pragma: no cover
        return hashlib.md5(  # type: ignore[call-arg]
            data, usedforsecurity=usedforsecurity
        ).hexdigest()

except TypeError:  # pragma: no cover

    def md5_hexdigest(data: bytes, *, usedforsecurity: bool = True) -> str:
        return hashlib.md5(data).hexdigest()

# Compatibility wrapper for `mimetypes.guess_type` to support `os.PathLike` on <py3.8
def guess_type(
    path: typing.Union[str, "os.PathLike[str]"], strict: bool = True
) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
    if sys.version_info < (3, 8):  # pragma: no cover
        path = os.fspath(path)
    return mimetypes_guess_type(path, strict)



def get_file_info(path: typing.Union[str, "os.PathLike[str]"]) -> typing.Dict[str, typing.Any]:
    if sys.version_info < (3, 8):  # pragma: no cover
        path = os.fspath(path)
    stat_result = os.stat(path)
    last_modified = formatdate(stat_result.st_mtime, usegmt=True)
    etag_base = f"{str(stat_result.st_mtime)}-{str(stat_result.st_size)}"
    etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)
    return {
        "size": stat_result.st_size,
        "last-modified": last_modified,
        "etag": etag,
        "media-type": guess_type(path)[0] or "text/plain",
        "filename": os.path.basename(path),
    }


"""
Because FSSpec is inconsistent in name mapping, we gotta do this.
"""

_ASYNC_METHODS = [
    "async_call",
    "async_cat",
    "async_cat_file",
    "async_cat_ranges",
    "async_copy",
    "async_cp_file",
    "async_du",
    "async_exists",
    "async_expand_path",
    "async_find",
    "async_get",
    "async_get_file",
    "async_getxattr",
    "async_glob",
    "async_info",
    "async_isdir",
    "async_isfile",
    "async_list_buckets",
    "async_list_objects",
    "async_ls",
    "async_makedirs",
    "async_merge",
    "async_mkdir",
    "async_pipe",
    "async_pipe_file",
    "async_process_limits",
    "async_put",
    "async_put_file",
    "async_rm",
    "async_rm_file",
    "async_rmdir",
    "async_rm_files",
    "async_setxattrs",
    "async_size",
    "async_sizes",
]

_ASYNC_MAPPING_GS = {
    "_call": "async_call",
    "_cat": "async_cat",
    "_cat_file": "async_cat_file",
    "_cat_ranges": "async_cat_ranges",
    "_copy": "async_copy",
    "_cp_file": "async_cp_file",
    "_du": "async_du",
    "_exists": "async_exists",
    "_expand_path": "async_expand_path",
    "_find": "async_find",
    "_get": "async_get",
    "_get_file": "async_get_file",
    "_getxattr": "async_getxattr",
    "_glob": "async_glob",
    "_info": "async_info",
    "_isdir": "async_isdir",
    "_isfile": "async_isfile",
    "_list_buckets": "async_list_buckets",
    "_list_objects": "async_list_objects",
    "_ls": "async_ls",
    #"_makedir": "async_makedir",
    "_makedirs": "async_makedirs",
    "_merge": "async_merge",
    "_mkdir": "async_mkdir",
    "_pipe": "async_pipe",
    "_pipe_file": "async_pipe_file",
    "_process_limits": "async_process_limits",
    "_put": "async_put",
    "_put_file": "async_put_file",
    "_rm": "async_rm",
    "_rm_file": "async_rm_file",
    "_rmdir": "async_rmdir",
    "_rm_files": "async_rm_files",
    "_setxattrs": "async_setxattrs",
    "_size": "async_size",
    "_sizes": "async_sizes",
}


_ASYNC_MAPPING_S3 = {
    "_call_s3": "async_call",
    "_cat": "async_cat",
    "_cat_file": "async_cat_file",
    "_cat_ranges": "async_cat_ranges",
    "_copy": "async_copy",
    "_cp_file": "async_cp_file",
    "_du": "async_du",
    "_exists": "async_exists",
    "_expand_path": "async_expand_path",
    "_find": "async_find",
    "_get": "async_get",
    "_get_file": "async_get_file",
    "_getxattr": "async_getxattr",
    "_glob": "async_glob",
    "_info": "async_info",
    "_isdir": "async_isdir",
    "_isfile": "async_isfile",
    "_lsbuckets": "async_list_buckets",
    "_lsdir": "async_list_objects",
    "_ls": "async_ls",
    "_makedirs": "async_makedirs",
    "_merge": "async_merge",
    "_mkdir": "async_mkdir",
    "_pipe": "async_pipe",
    "_pipe_file": "async_pipe_file",
    "_process_limits": "async_process_limits",
    "_put": "async_put",
    "_put_file": "async_put_file",
    "_rm": "async_rm",
    "_rm_file": "async_rm_file",
    "_rmdir": "async_rmdir",
    "_bulk_delete": "async_rm_files",
    "_setxattr": "async_setxattrs",
    "_size": "async_size",
    "_sizes": "async_sizes",
}

_ASYNC_SYNTAX_MAPPING = {
    'gs': _ASYNC_MAPPING_GS,
    's3': _ASYNC_MAPPING_S3,
    'minio': _ASYNC_MAPPING_S3,
}




"""
Cloud Provider Configs
"""

import os
import json
import pathlib
import multiprocessing as mp
from typing import Optional, Union, Any, Dict, TYPE_CHECKING

from pydantic import Json
from loguru import logger

from ..types.config import ConfigModel


try:
    from google.colab import drive
    _is_colab = True
except ImportError: _is_colab = False

class CloudConfig(ConfigModel):
    config_dir: Optional[pathlib.Path] = pathlib.Path("~/.auth").resolve()
    boto_config: Optional[pathlib.Path] = pathlib.Path("~/.boto")
    
    """ 
    AWS Specific 
    """
    aws_access_token: Optional[str] = ""
    aws_access_key_id: Optional[str] = ""
    aws_secret_access_key: Optional[str] = ""
    aws_region: Optional[str] = "us-east-1"
    set_s3_endpoint: Optional[bool] = True
    s3_config: Optional[Json] = None

    """ 
    GCP Specific 
    """

    gcp_project: Optional[str] = ""
    gcloud_project: Optional[str] = ""
    google_cloud_project: Optional[str] = ""
    google_application_credentials: Optional[pathlib.Path] = pathlib.Path("~/adc.json").resolve()

    gcs_client_config: Optional[Json] = None
    gcs_config: Optional[Json] = None

    """
    Minio Specific
    """
    minio_endpoint: Optional[str] = ""
    minio_access_key: Optional[str] = ""
    minio_secret_key: Optional[str] = ""
    minio_access_token: Optional[str] = ""
    minio_config: Optional[Json] = None
    minio_signature_ver: str = 's3v4'

    """
    S3-Compatiable Generic
    """
    s3compat_endpoint: Optional[str] = ""
    s3compat_region: Optional[str] = ""
    s3compat_access_key: Optional[str] = ""
    s3compat_secret_key: Optional[str] = ""
    s3compat_access_token: Optional[str] = ""
    s3compat_config: Optional[Json] = None

    max_workers: Optional[int] = mp.cpu_count() * 2

    def create_adc(self, data: Union[str, Dict[str, Any]], path: str = None):
        """
        Create a new ADC based on the passed data and writes it to 
        path or GOOGLE_APPLICATION_CREDENTIALS
        """
        if isinstance(data, str): data = json.loads(data)
        path: pathlib.Path = pathlib.Path(path) if path else self.google_application_credentials
        path.write_text(json.dumps(data, indent = 2, ensure_ascii=False))
        self.google_application_credentials = path
        

    def get_gcp_project(self):
        for v in [self.gcp_project, self.gcloud_project, self.google_cloud_project]:
            if v: return v

    def get_boto_config_path(self):
        if not self.boto_config:
            self.boto_config = pathlib.Path('/root/.boto') if _is_colab else pathlib.Path('~/.boto').resolve(True)
        return self.boto_config

    def get_s3_endpoint(self):
        return f'https://s3.{self.aws_region}.amazonaws.com'

    def get_boto_path(self):
        boto_config = self.get_boto_config_path()
        if boto_config.exists(): return boto_config
        self.config_dir.mkdir(create_parents=True, exist_ok=True)
        return self.config_dir.joinpath('.boto')
    
    def should_write_boto(self):
        boto_config = self.get_boto_config_path()
        return not boto_config.exists()
 
    def get_boto_values(self):
        t = "[Credentials]\n"
        if self.aws_access_key_id:
            t += f"aws_access_key_id = {self.aws_access_key_id}\n"
            t += f"aws_secret_access_key = {self.aws_secret_access_key}\n"
        if self.google_application_credentials.exists():
            t += f"gs_service_key_file = {self.google_application_credentials.as_posix()}\n"
        t += "\n[Boto]\n"
        t += "https_validate_certificates = True\n"
        t += "\n[GSUtil]\n"
        t += "content_language = en\n"
        t += "default_api_version = 2\n"
        if gcp_project := self.get_gcp_project():
            t += f"default_project_id = {gcp_project}\n"
        return t


    def write_botofile(self, overwrite: bool = False, **kwargs):
        if self.should_write_boto:
            p = self.get_boto_path()
            if not p.exists() or overwrite:
                logger.info(f'Writing Botofile to {p.as_posix()}')
                p.write_text(self.get_boto_values())
            else: 
                logger.error(f'Botofile {p.as_posix()} exists and overwrite is False. Not overwriting')
            return p
        else: 
            logger.warning(f'Skipping writing Botofile as BotoConfig = {self.boto_config.as_posix()} exists')
            return self.boto_config


    def set_auth_env(self):
        if self.google_application_credentials.exists():
            os.system['GOOGLE_APPLICATION_CREDENTIALS'] = self.google_application_credentials.as_posix()
        gcp_project = self.get_gcp_project()
        if gcp_project: 
            os.system['GOOGLE_CLOUD_PROJECT'] = gcp_project

        try:
            botopath = self.get_boto_path()

            ## We know this is our custom botofile
            if botopath.exists() and self.should_write_boto():
                os.system['BOTO_PATH'] = botopath.as_posix()
                os.system['BOTO_CONFIG'] = botopath.as_posix()
        except: pass
        if self.aws_access_key_id:
            os.system['AWS_ACCESS_KEY_ID'] = self.aws_access_key_id
            os.system['AWS_SECRET_ACCESS_KEY'] = self.aws_secret_access_key
        
        if self.minio_endpoint:
            os.system['MINIO_ENDPOINT'] = self.minio_endpoint
        if self.set_s3_endpoint:
            os.system['S3_ENDPOINT'] = self.get_s3_endpoint()
        if self.minio_access_key:
            os.system['MINIO_ACCESS_KEY'] = self.minio_access_key
            os.system['MINIO_SECRET_KEY'] = self.minio_secret_key
        

    def update_auth(self, **config):
        self.update_config(**config)
        self.set_auth_env()
    
    class Config:
        env_prefix = ""
        


        